﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using SecsTranslatorEditor.SecsLanguage.Formatters;

namespace SecsTranslatorEditor.SecsLanguage.Manual
{
	/// <summary>
	/// Represents the secs language syntax processor.
	/// </summary>
	internal class SecsLanguageSyntax
	{
		#region Private Fields
		/// <summary>
		/// Contains the regex used to split text.
		/// </summary>
		private readonly Regex mSplitRegex =
			new Regex(@"(\s+)", RegexOptions.Multiline);

		/// <summary>
		/// Conatins all syntax elements that build the language.
		/// </summary>
		private IEnumerable<ISecsLanguageSyntaxElement> mSyntaxElements;
		#endregion

		#region Constructors
		/// <summary>
		/// Initializes a new instance of the <see cref="SecsLanguageSyntax"/> class.
		/// </summary>
		internal SecsLanguageSyntax()
		{
			BuildSyntax();
		}
		#endregion

		#region Public Properties
		/// <summary>
		/// Gets the regex used to split text in single words.
		/// </summary>
		public Regex SplitWordsRegex
		{
			get { return mSplitRegex; }
		}

		/// <summary>
		/// Gets the syntax keywords.
		/// </summary>
		internal IEnumerable<ISecsLanguageSyntaxElement> SyntaxElements => mSyntaxElements;
		#endregion

		#region Public Methods
		/// <summary>
		/// Gets the type of the language word.
		/// </summary>
		/// <param name="word">The given language word.</param>
		/// <returns>Returns the the type of the word.</returns>
		public ISecsLanguageSyntaxElement GetLanguageItemType(string word)
		{
			return mSyntaxElements.FirstOrDefault(x => x.IsMatch(word));
		}
		#endregion

		#region Private Methods
		/// <summary>
		/// Builds the standard SECS language syntax elements.
		/// </summary>
		private void BuildSyntax()
		{
			mSyntaxElements = new List<ISecsLanguageSyntaxElement>()
			{
				// Secs translator reserved keywoards
				new SecsLanguageKeyword("Base", new BaseKeywordFormatter()),
				new SecsLanguageKeyword("Using", new BaseKeywordFormatter()),
				new SecsLanguageKeyword("Request"),
				new SecsLanguageKeyword("Response"),
				new SecsLanguageKeyword("SecsItem"),
				new SecsLanguageKeyword("Step"),
				new SecsLanguageKeyword("foreach"),
				new SecsLanguageKeyword("Sequence"),
				new SecsLanguageKeyword("override"),
				new SecsLanguageKeyword("stoponerror"),

				// Standard .NET reserved type names
				new SecsLanguageKeyword("string"),
				new SecsLanguageKeyword("int"),
				new SecsLanguageKeyword("decimal"),
				new SecsLanguageKeyword("bool"),
				new SecsLanguageKeyword("single"),
				new SecsLanguageKeyword("char"),
				new SecsLanguageKeyword("byte"),
				new SecsLanguageKeyword("float"),
				new SecsLanguageKeyword("object")
			};
		}
		#endregion
	}
}
