﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Input;
using System.Windows.Media;
using Infragistics.Controls.Editors;
using Infragistics.Documents;
using Infragistics.Documents.Parsing;
using SecsTranslatorEditor.Helpers;
using SecsTranslatorEditor.SecsLanguage.Manual;

namespace SecsTranslatorEditor.SecsLanguage
{

	/// <summary>
	/// Custom text box object with two additional dependency properties, scroll and caret position to assist with
	/// interaction between the view and view model.
	/// </summary>
	public class SecsTranslatorEditor : XamSyntaxEditor
	{
		#region Dependency Properties
		/// <summary>View dependency property for binding the caret position in the text editor.</summary>
		private static readonly DependencyProperty sCaretPositionExternalProperty =
		SecsTranslatorEditorDependencyProperties.RegisterCaretPositionExternal(
		(sender, args) => ((SecsTranslatorEditor)sender).CaretPositionExternal = (int)args.NewValue);

		/// <summary>View dependency property for binding the editor's read-only behavior.</summary>
		private static readonly DependencyProperty sIsReadOnly =
		SecsTranslatorEditorDependencyProperties.RegisterIsReadOnly(
		(target, newValue) => target?.HandleIsReadOnlyChanged(newValue));

		/// <summary>View dependency property for binding the raw text content of the editor.</summary>
		private static readonly DependencyProperty sTextProperty = SecsTranslatorEditorDependencyProperties
		.RegisterText(new UIPropertyMetadata(null, OnScriptBodyChanged));

		/// <summary>View dependency property for binding the search pattern for searching the content.</summary>
		private static readonly DependencyProperty sSearchPatternProperty =
		SecsTranslatorEditorDependencyProperties.RegisterSearchPattern(
		(target, newValue) => target?.HandleSearchPatternChanged(newValue));

		/// <summary>View dependency property for binding the search results.</summary>
		private static readonly DependencyProperty sSearchResultsProperty =
		SecsTranslatorEditorDependencyProperties.RegisterSearchResults();

		/// <summary>View dependency property for binding find item behavior.</summary>
		private static readonly DependencyProperty sFindItemProperty =
		SecsTranslatorEditorDependencyProperties.RegisterFindItem(
		(target, newValue) => target?.HandleFoundItemChanged(newValue));

		private static readonly DependencyProperty sSelectTextPosition =
		SecsTranslatorEditorDependencyProperties.RegisterSelectTextPosition();

		private static readonly DependencyProperty sTextChangedCommandProperty =
		SecsTranslatorEditorDependencyProperties.RegisterTextChangedCommand();

		#endregion

		#region Private Variables

		private readonly ObservableCollection<FindResultInfo> mSearchResults = new ObservableCollection<FindResultInfo>();

		private string mText;
		private bool mIsTextChanged;
		#endregion

		#region Constructors
		/// <summary>Initializes a new instance of the SecsTranslatorEditor class.</summary>
		public SecsTranslatorEditor()
		{
			FontFamily = new FontFamily("Consolas");
			FontSize = 12;
			Document = new TextDocument { IsReadOnly = false };
			Document.TextChanged += OnTextChanged;

			SelectionChanged += OnSelectionChanged;
			Document.Language = InitializeLanguage();

			SearchResults = new ObservableCollection<FindResultInfo>(mSearchResults);
			mSearchResults.CollectionChanged += OnSearchResultsChanged;

			Loaded += OnEditorLoaded;
			ActiveDocumentViewChanged += OnActiveDocumentChanged;
		}
		#endregion

		#region Public Properties
		/// <summary>Gets or sets the caret position in the textbook.</summary>
		//[UsedImplicitly]
		public int CaretPositionExternal
		{
			get { return (int)GetValue(sCaretPositionExternalProperty); }
			set { UpdateCaretPositionFromExternal(value); }
		}

		/// <summary>Gets or sets the command corresponding to a text changed.</summary>
		//[UsedImplicitly]
		public ICommand TextChangedCommand
		{
			get { return (ICommand)GetValue(sTextChangedCommandProperty); }
			set { SetValue(sTextChangedCommandProperty, value); }
		}

		/// <summary>Gets or sets a value indicating the latest value of status bar.</summary>
		/// <value>The value indicating the line number and character number that carat seats.</value>
		[DefaultValue("Line:0, Character:0")]
		//[UsedImplicitly]
		public string SelectTextPosition
		{
			get { return (string)GetValue(sSelectTextPosition); }
			set { SetValue(sSelectTextPosition, value); }
		}

		/// <summary>Gets or sets a value indicating whether the text displayed by editor is read-only.</summary>
		/// <value>The value indicating whether the text is read-only.</value>
		[DefaultValue(false)]
		//[UsedImplicitly]
		public bool IsReadOnly
		{
			get { return (bool)GetValue(sIsReadOnly); }
			set { SetValue(sIsReadOnly, value); }
		}

		/// <summary>Gets or sets the a search pattern which can be used to search for in the displayed text.</summary>
		/// <value>The search pattern.</value>
		[DefaultValue("")]
		//[UsedImplicitly]
		public string SearchPattern
		{
			get { return (string)GetValue(sSearchPatternProperty); }
			set { SetValue(sSearchPatternProperty, value); }
		}

		/// <summary>
		/// Gets or sets available items that match the <see cref="SearchPattern"/> in the document of this editor.
		/// </summary>
		public ObservableCollection<FindResultInfo> SearchResults
		{
			get { return (ObservableCollection<FindResultInfo>)GetValue(sSearchResultsProperty); }
			set { SetValue(sSearchResultsProperty, value); }
		}

		/// <summary>Gets or sets the text that should be displayed in the document.</summary>
		//[UsedImplicitly]
		public string Text
		{
			get { return (string)GetValue(sTextProperty); }
			set { SetValue(sTextProperty, value); }
		}

		/// <summary>Gets or sets the item which should be found in the editor.</summary>
		//[UsedImplicitly]
		public FindResultInfo FindItem
		{
			get { return (FindResultInfo)GetValue(sFindItemProperty); }
			set { SetValue(sFindItemProperty, value); }
		}
		#endregion

		#region Private Methods
		private static void OnScriptBodyChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
		{
			if (d is SecsTranslatorEditor)
				(d as SecsTranslatorEditor).HandleTextChanged(e.NewValue);
		}

		private void UpdateCaretPositionFromExternal(int value)
		{
			Focus();
			ActiveDocumentView.GoToLineNumber(value + 1);
		}

		private void OnActiveDocumentChanged(object sender, ActiveDocumentViewChangedEventArgs e)
		{
			HandleTextChanged(Text);
		}

		private void OnTextChanged(object sender, TextChangedEventArgs e)
		{
			mIsTextChanged = true;
			SetValue(sTextProperty, e.After.GetText());
			TextChangedCommand?.Execute(e);
		}

		private void HandleTextChanged(object value)
		{
			try
			{
				// Assures that the provided new configuration text is different than the current one.
				if (value == null || value.ToString() == mText)
					return;

				string newText = value.ToString();

				if (mIsTextChanged)
				{
					// Doesn't reinitialize if the configuration has been changed directly by the raw editor.
					mIsTextChanged = false;
					mText = newText;
				}
				else
				{
					// Reinitializes the text document if the configuration has been changed 
					// by a graphical editor or it is a first-time initialization. 
					if (ActiveDocumentView != null && ActiveDocumentView.Document != null)
					{
						ActiveDocumentView.Document.InitializeText(newText);
						mText = newText;
					}
				}
			}
			catch (Exception ex)
			{
				ErrorHandler.OnError(ex, "Unexpected error while handling text change in XmlEditor.");
			}
		}

		private CustomLanguage InitializeLanguage()
		{
			var grammar = PlainTextLanguage.Instance.Grammar.Clone();
			new SecsLanguageSyntax().SyntaxElements.Cast<SecsLanguageKeyword>().ToList().ForEach(element =>
			{
				grammar.LexerStates.DefaultLexerState.Symbols.Insert(0, element.Keyword, element.Keyword,
					TerminalSymbolComparison.LiteralIgnoreCase).LanguageElement = LanguageElement.ContextSensitiveKeyword;
			});
			return new CustomLanguage(grammar);
		}

		private void OnEditorLoaded(object sender, RoutedEventArgs e)
		{
			Document.IsReadOnly = IsReadOnly;
		}

		private void OnSelectionChanged(object sender, SyntaxEditorSelectionChangedEventArgs e)
		{
			try
			{
				int line = e.DocumentView.SelectionManager.SelectionStartNormalized.GetTextLocation().Line;
				int character = e.DocumentView.SelectionManager.SelectionStartNormalized.GetTextLocation().Character;
				SelectTextPosition = $"Line:{(line + 1).ToString()}, Character:{(character + 1).ToString()}";

				if (FindItem == null || !(sender is SecsTranslatorEditor))
					return;

				Focus();
			}
			catch (Exception ex)
			{
				ErrorHandler.OnError(ex, "Unexpected error happened on text selection changed in XmlEditor.");
			}
		}

		private void OnSearchResultsChanged(object sender, NotifyCollectionChangedEventArgs e)
		{
			try
			{
				if (SearchResults == null)
					SearchResults = new ObservableCollection<FindResultInfo>();

				lock (SearchResults)
				{
					if (e.Action == System.Collections.Specialized.NotifyCollectionChangedAction.Add)
					{
						foreach (FindResultInfo item in e.NewItems)
							SearchResults.Add(item);
					}
					else
						SearchResults.Clear();
				}
			}
			catch (Exception ex)
			{
				ErrorHandler.OnError( ex, "Unexpected error happened on search results changed in XmlEditor.");
			}
		}

		private void HandleSearchPatternChanged(object value)
		{
			try
			{
				if (string.IsNullOrEmpty(value?.ToString()))
					return;

				var data = value.ToString();
				var tsc = new TextSearchCriteria(data);
				var searchResults = ActiveDocumentView.Document.CurrentSnapshot.FindAll(tsc);

				if (searchResults == null)
					return;

				// Clear the current results.
				if (mSearchResults.Count > 0)
					mSearchResults.Clear();

				if (searchResults.Results.Count == 0)
				{
					DialogManager.ShowMessageBox("There are no items matching your search criteria.", "Search results");
					return;
				}

				foreach (var searchResult in searchResults.Results)
				{
					// Obtain the span containing the search result
					var span = searchResult.SpanFound;

					// Obtain the line at which the result is found
					var lineInfo = span.Snapshot.LineFromOffset(span.Offset);

					mSearchResults.Add(new FindResultInfo
					{
						Row = lineInfo.LineNumber,
						Text = lineInfo.GetText(false).Trim(),
						Offset = span.Offset,
						Pattern = data
					});
				}
			}
			catch (Exception ex)
			{
				ErrorHandler.OnError(ex,
					"Unexpected error happened while handling search pattern change in XmlEditor.");
			}
		}

		private void HandleIsReadOnlyChanged(object value)
		{
			if (ActiveDocumentView?.Document != null)
				ActiveDocumentView.Document.IsReadOnly = Convert.ToBoolean(value);
		}

		private void HandleFoundItemChanged(object value)
		{
			try
			{
				var findResult = value as FindResultInfo;
				if (findResult == null || FindResultUpdateNotRequired())
					return;

				var currentSnapshot = ActiveDocumentView.CurrentSnapshot;

				var start = findResult.Offset;
				var end = start + findResult.Pattern.Length;
				var stl = currentSnapshot.LocationFromOffset(start);
				var etl = currentSnapshot.LocationFromOffset(end);

				var ssp = SnapshotPoint.FromTextLocation(stl, currentSnapshot);
				var esp = SnapshotPoint.FromTextLocation(etl, currentSnapshot);

				ActiveDocumentView.SelectionManager.SelectionStart = ssp;
				ActiveDocumentView.SelectionManager.SelectionEnd = esp;

				// Make sure line is visible in editor and not scrolled away.
				ActiveDocumentView.BringLineIntoView(stl.Line, true);
			}
			catch (Exception ex)
			{
				ErrorHandler.OnError(ex,
					"Unexpected error happened while handling found item change in XmlEditor.");
			}
		}

		private bool FindResultUpdateNotRequired()
		{
			return ActiveDocumentView?.CurrentSnapshot == null || ActiveDocumentView.SelectionManager == null;
		}
		#endregion
	}
}
